from typing import Union
from datetime import datetime
from collections import defaultdict

from torch.utils.tensorboard import SummaryWriter

import os
import json
import time

class TrainingStats:
    def __init__(self,
                 metrics_tracking=False,
                 log_dir: Union[str or os.PathLike] = f'trainer_output_{datetime.now().strftime("%H%M%S_%d%h%Y")}'):
        default_stats = {
            # Training progress
            'curr_epoch': 0,
            'curr_global_step': 0,
            'max_steps': 0,
            'train_examples': 0,
            'train_epochs': 0,
            'train_batch_size': 0,
            'train_metrics': defaultdict(list) if metrics_tracking else {},
            'closed_reason': None,
            # Evaluate progress
            'eval_examples': 0,
            'eval_batch_size': 0,
            'total_eval_steps': 0,
            'evaluation_strategy': None,
            'metric_for_best_model': 'loss',
            'eval_metrics': defaultdict(list) if metrics_tracking else {}

        }
        self._stats = default_stats
        self.metrics_tracking = metrics_tracking
        if metrics_tracking:
            self.tracking_board = SummaryWriter(log_dir=os.path.join(log_dir, 'logs'))
        self.cached_time = time.time()

    def set_stats(self, stats):
        self._stats = stats

    def get_value(self, key, default=None):
        return self._stats.get(key, default)

    def set_value(self, key, value):
        self._stats[key] = value

    def inc_value(self, key, count=1, start=0):
        d = self._stats
        if d.setdefault(key, start) is None:
            d[key] = count
        else:
            d[key] = d.setdefault(key, start) + count

    def update_time(self):
        self.cached_time = time.time()

    def clear_stats(self):
        self._stats.clear()

    def save_to_json(self, json_path: str):
        """Save the content of this instance in JSON format inside `json_path`."""
        json_string = json.dumps(self._stats, indent=2, sort_keys=True) + "\n"
        with open(json_path, "w", encoding="utf-8") as f:
            f.write(json_string)

    def load_from_json(self, json_path: str):
        """Create an instance from the content of `json_path`."""
        with open(json_path, "r", encoding="utf-8") as f:
            text = f.read()
        self._stats = json.loads(text)

    def get_evaluation_step(self):
        if self._stats['evaluation_strategy'] == 'epoch':
            return self.get_value('curr_epoch')
        else:
            return self.get_value('curr_global_step')

    def _update_metrics(self, metrics: dict, metrics_name: str):
        if self.metrics_tracking:
            for k, v in metrics.items():
                self.tracking_board.add_scalar(tag=k,
                                               scalar_value=v,
                                               global_step=self.get_evaluation_step())
                self._stats[metrics_name][k].append(v)
        else:
            self.set_value(metrics_name, metrics)

    def _metrics_to_str(self, metrics_name):
        metrics_str = ''
        for k, v in self._stats[metrics_name].items():
            curr_value = v[-1] if self.metrics_tracking else v
            metrics_str += f'{k} = {curr_value:.4f}; '
        return metrics_str

    def update_train_metrics(self, metrics: dict):
        self._update_metrics(metrics, metrics_name='train_metrics')

    def update_eval_metrics(self, metrics: dict):
        self._update_metrics(metrics, metrics_name='eval_metrics')

    def get_train_metrics_str(self) -> str:
        return self._metrics_to_str(metrics_name='train_metrics')

    def get_eval_metrics_str(self):
        return self._metrics_to_str(metrics_name='eval_metrics')



class EarlyStopping:
    def __init__(self,
                 patience: int = 3,
                 min_delta: float = 0.0):
        self.patience = patience
        self.wait_count = 0
        self.min_delta = min_delta
        self.best_score = None

    def __call__(self, curr_score, *args, **kwargs):
        if abs(self.best_score - curr_score) > self.min_delta:
            self.wait_count += 1
            if self.wait_count >= self.patience:
                return True
        return False

    def restart(self, best_score):
        self.wait_count = 0
        self.best_score = best_score



from typing import Optional, Any, List, Dict, Callable, TypeVar
from tqdm import tqdm

from kheiron.utils import *
from kheiron.constants import SavedFiles
from kheiron.opts import TrainingOptions
from kheiron.early_stop import EarlyStopping

from torch import nn
from torch.utils.data import Dataset as TorchDataset, DataLoader, RandomSampler
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR

from datasets.arrow_dataset import Dataset
import os
import math
import torch
import time
import inspect

T = TypeVar('T')


class Trainer:
    def __init__(self,
                 model: nn.Module,
                 opts: TrainingOptions,
                 train_set: Optional[Dataset] = None,
                 eval_set: Optional[Dataset] = None,
                 optimizer: Optional[Optimizer] = None,
                 lr_scheduler: Optional[LambdaLR] = None,
                 eval_metrics: Optional[DefaultMetrics] = None,
                 collate_fn:  Callable[[List[T]], Any] = None):
        self.opts = opts
        self.optim = optimizer
        self.scheduler = lr_scheduler
        self.eval_metrics = eval_metrics
        self.model = model
        self.train_set = train_set
        self.eval_set = eval_set
        self.collate_fn = collate_fn
        self.early_stopping = None

        # Create output directory if not exist
        if not os.path.exists(self.opts.output_dir):
            os.makedirs(self.opts.output_dir)

        # Create training statistics
        self.stats = TrainingStats(metrics_tracking=opts.track_metrics, log_dir=opts.output_dir)

        # Set seed
        set_ramdom_seed(self.opts.seed)

        # Move model to device
        self.model = model.to(self.opts.device)

        # Data unused features
        self.signature_features = ["label", "label_ids"] + list(inspect.signature(self.model.forward).parameters.keys())
        self.ununsed_features = []

        # Set up the optimizer
        if self.optim is None:
            self.create_optimizer()

        # Set up the Scheduler
            self.scheduler_update_strategy = 'step'

        if self.eval_metrics is None:
            self.eval_metrics = DefaultMetrics()

        if self.opts.early_stopping_steps > 0:
            self.early_stopping = EarlyStopping(patience=self.opts.early_stopping_steps,
                                                min_delta=self.opts.early_stopping_min_delta)

    def _close_train_progress(self):
        LOGGER.info('\n'
                    f'Trainer closed ({self.stats.get_value("closed_reason")})\n'
                    f'Total progress spend time: {get_spend_time(self.stats.cached_time, time.time())}\n'
                    f'Trained epoch {self.stats.get_value("curr_epoch")}/{self.stats.get_value("train_epochs")}; '
                    f'Global step = {self.stats.get_value("curr_global_step")}/{self.stats.get_value("max_steps")}\n' \
                    f'Best model at {self.stats.get_value("evaluation_strategy")} {self.stats.get_value("best_step")}: ' \
                    f'eval_loss = {self.stats.get_value("best_loss")}; ' \
                    f'key metric (eval_{self.stats.get_value("metric_for_best_model")}) = {self.stats.get_value("best_score")} \n' \
                    f'Model checkpoint and logs saved at `{self.opts.output_dir}`')

    def _remove_ununsed_features(self, feature: dict):
        for k in self.ununsed_features:
            del feature[k]

    def _wrap_collator_with_removed_unused_features(self, features: List[dict]):
        if len(self.ununsed_features) == 0 and isinstance(features, dict):
            self.ununsed_features = list(set(features.keys()) - set(self.signature_features))
        for feature in features:
            self._remove_ununsed_features(feature)
        if self.collate_fn is None:
            return features
        return self.collate_fn(features)

    def _save_training_step(self):
        # Save training arguments
        torch.save(self.opts, os.path.join(self.opts.output_dir, SavedFiles.TRAINING_ARGS_FILE))

        # Save training stats
        self.stats.save_to_json(os.path.join(self.opts.output_dir, SavedFiles.TRAINING_STATS_FILE))

        # Save optimizer
        torch.save(self.optim.state_dict(), os.path.join(self.opts.output_dir, SavedFiles.OPTIMIZER_FILE))

        # Save Scheduler
        torch.save(self.scheduler.state_dict(), os.path.join(self.opts.output_dir, SavedFiles.SCHEDULER_FILE))

    def _save_model(self):
        state_dict = self.model.state_dict()
        LOGGER.info(f"Saving model checkpoint to {os.path.join(self.opts.output_dir, SavedFiles.WEIGHT_MODEL_FILE)}")
        torch.save(state_dict, os.path.join(self.opts.output_dir, SavedFiles.WEIGHT_MODEL_FILE))

    def _get_optimizer_parameter_names(self, model: nn.Module, parent_name: str = '') -> List[str]:
        if parent_name in self.opts.no_decay_param_names:
            return []
        param_names = []
        for model_name, child_model in model.named_children():
            if model_name not in self.opts.no_decay_param_names:
                param_names += [
                    f'{model_name}.{child_name}'
                    for child_name in self._get_optimizer_parameter_names(child_model, f'{parent_name}.{model_name}')
                ]
        param_names += [n for n in list(model._parameters.keys()) if n not in self.opts.no_decay_param_names]
        return param_names

    def _preprocess_inputs(self, batch: dict):
        for k, v in batch.items():
            batch[k] = v.to(self.opts.device)

    def _preprocess_logits(self, logits):
        if self.opts.task == 'text-classification':
            return torch.argmax(logits, dim=-1)

    def _fit_one_epoch(self, process_bar):
        start_time = time.time()
        self.model.train()
        ep_loss = torch.tensor(0.0).to(self.opts.device)
        len_iterator = len(process_bar)

        for step_batch in process_bar:
            process_bar.desc = f'  [TRAIN] Epoch = ' \
                               f'{self.stats.get_value("curr_epoch")}/{self.stats.get_value("train_epochs")}; ' \
                               f'Global step = {self.stats.get_value("curr_global_step")}/' \
                               f'{self.stats.get_value("max_steps")}; ' \
                               f'Learning rate = {get_lr_strings(self.optim)}'
            self.model.zero_grad()
            self._preprocess_inputs(step_batch)
            outputs = self.model(**step_batch)
            if isinstance(outputs, dict) and 'loss' not in outputs:
                raise ValueError(
                    f"The 'loss' key was not found in the returned model results, only the following keys: "
                    f"{outputs.keys()}."
                )
            step_loss = outputs['loss'] if isinstance(outputs, dict) else outputs[0]
            if self.opts.gradient_accumulation_steps > 1:
                step_loss = step_loss / self.opts.gradient_accumulation_steps
            step_loss.backward()
            ep_loss += step_loss.detach()
            torch.nn.utils.clip_grad_norm_(parameters=self.model.parameters(), max_norm=self.opts.max_grad_norm)
            self.optim.step()
            if self.scheduler_update_strategy == 'step':
                self.scheduler.step()
            self.stats.inc_value('curr_global_step')

            # Check evaluation strategy
            if self.opts.evaluation_strategy == 'step' and \
                    self.stats.get_value('curr_global_step') % self.opts.eval_steps == 0:
                eval_metrics = self.evaluate()
                if self._is_better_model(eval_metrics, f'eval_{self.opts.metric_for_best_model}'):
                    self._save_model()
                    self._save_training_step()

                    # Update statistics for best model
                    self.stats.set_value('best_loss', eval_metrics['eval_loss'])
                    self.stats.set_value('best_score', eval_metrics[f'eval_{self.opts.metric_for_best_model}'])
                    self.stats.set_value('best_step', self.stats.get_evaluation_step())

        metrics = {'train_loss': ep_loss.item()/len_iterator}

        # Update statistics with evaluation scores
        self.stats.update_train_metrics(metrics)

        # Log result
        LOGGER.info(f'\n'
                    f'  Fit epoch results: \n'
                    f'    Key metrics: loss= {metrics["train_loss"]:.4f}\n'
                    f'    Full metrics: {self.stats.get_train_metrics_str()}')

        return metrics

    def _is_better_model(self, metrics: Dict, key_metric: str) -> bool:
        if self.opts.greater_is_better \
                and self.stats.get_value('best_score', float('-inf')) < metrics[key_metric]:
            return True
        elif not self.opts.greater_is_better \
                and self.stats.get_value('best_score', float('inf')) > metrics[key_metric]:
            return True
        else:
            return False

    def create_optimizer(self):
        decay_param_names = self._get_optimizer_parameter_names(self.model)
        optimizer_grouped_parameters = [
            {
                'params': [p for n, p in self.model.named_parameters() if n in decay_param_names],
                'lr': self.opts.learning_rate, "weight_decay": self.opts.weight_decay,
            },
            {
                'params': [p for n, p in self.model.named_parameters() if n not in decay_param_names],
                'lr': self.opts.learning_rate, "weight_decay": 0.0,
            }
        ]
        if self.opts.optimizer_name == OptimizerNames.AdamW['name']:
            self.optim = OptimizerNames.AdamW['cls'](optimizer_grouped_parameters,
                                                     lr=self.opts.learning_rate,
                                                     betas=(self.opts.adam_beta1, self.opts.adam_beta2),
                                                     eps=self.opts.adam_epsilon)
        if self.opts.optimizer_name == OptimizerNames.SGD['name']:
            self.optim = OptimizerNames.SGD['cls'](optimizer_grouped_parameters,
                                                   lr=self.opts.learning_rate,
                                                   momentum=self.opts.momentum
                                                   )

    def create_scheduler(self, num_training_steps: int):
        warmup_steps = self.opts.warmup_steps if self.opts.warmup_steps > 0 \
            else math.ceil(num_training_steps * self.opts.warmup_proportion)
        training_steps = self.opts.epochs * num_training_steps
        if self.opts.scheduler_name == SchedulerNames.Linear['name']:
            self.scheduler = SchedulerNames.Linear['cls'](optimizer=self.optim,
                                                          num_warmup_steps=warmup_steps,
                                                          num_training_steps=training_steps)
            self.scheduler_update_strategy = SchedulerNames.Linear['update_strategy']
        if self.opts.scheduler_name == SchedulerNames.StepLR['name']:
            self.scheduler = SchedulerNames.StepLR['cls'](optimizer=self.optim,
                                                          step_size=self.opts.decay_step,
                                                          gamma=self.opts.gamma,
                                                          last_epoch=-1)
            self.scheduler_update_strategy = SchedulerNames.StepLR['update_strategy']

    def evaluate(self):
        if self.eval_set is None or \
                (not isinstance(self.eval_set, Dataset) and not isinstance(self.eval_set, TorchDataset)):
            raise TypeError(
                f"Expected type `torch.utils.data.Dataset` or `datasets.arrow_dataset.Dataset` for `eval_set`, "
                f"got '{type(self.eval_set)}' instead."
            )
        eval_iterator = DataLoader(self.eval_set,
                                   batch_size=self.opts.eval_batch_size,
                                   collate_fn=self._wrap_collator_with_removed_unused_features,
                                   num_workers=self.opts.processor_num_workers)
        num_examples = len(self.eval_set)
        len_iterator = len(eval_iterator)

        # Update statistics
        self.stats.set_value('eval_examples', num_examples)
        self.stats.set_value('eval_batch_size', self.opts.eval_batch_size)
        self.stats.set_value('total_eval_steps', len_iterator)

        eval_progress_bar = tqdm(eval_iterator,
                                 total=len_iterator,
                                 leave=True,
                                 position=0,
                                 desc=f'  [EVAL] Epoch = {self.stats.get_value("curr_epoch")}/'
                                      f'{self.stats.get_value("train_epochs")}; '
                                      f'Global step = {self.stats.get_value("curr_global_step")}/'
                                      f'{self.stats.get_value("max_steps")}')
        self.model.eval()
        eval_loss = torch.tensor(0.0).to(self.opts.device)
        golds, preds = [], []
        with torch.no_grad():
            for step_batch in eval_progress_bar:
                self._preprocess_inputs(step_batch)
                outputs = self.model(**step_batch)
                if isinstance(outputs, dict) and 'loss' not in outputs:
                    raise ValueError(
                        f"The 'loss' key was not found in the returned model results, only the following keys: "
                        f"{outputs.keys()}."
                    )
                step_loss = outputs['loss'] if isinstance(outputs, dict) else outputs[0]
                eval_loss += step_loss.detach()
                if 'labels' in step_batch:
                    golds.extend(step_batch['labels'].detach().cpu().tolist())
                if isinstance(outputs, dict) and 'logits' in outputs:
                    logits = self._preprocess_logits(outputs['logits'])
                    preds.extend(logits.detach().cpu().tolist())

        scores = self.eval_metrics(true_seqs=golds, pred_seqs=preds, task=self.opts.task)
        metrics = {f"eval_{key}": score for key, score in scores.items()}
        metrics['eval_loss'] = eval_loss.item() / len_iterator

        # Update statistics with evaluation scores
        self.stats.update_eval_metrics(metrics)

        # Log result
        LOGGER.info(f'\n'
                    f'  Evaluation results:\n'
                    f'     Key metrics: loss: {metrics["eval_loss"]:.4f}; '
                    f'{self.opts.metric_for_best_model}: {metrics[f"eval_{self.opts.metric_for_best_model}"]:.4f}\n'
                    f'     Full metrics: {self.stats.get_eval_metrics_str()}')
        return metrics

    def train(self):
        self.stats.update_time()

        if self.train_set is None or \
                (not isinstance(self.train_set, Dataset) and not isinstance(self.train_set, TorchDataset)):
            raise TypeError(
                f"Expected type `torch.utils.data.Dataset` or `datasets.arrow_dataset.Dataset` for `train_set`, "
                f"got '{type(self.train_set)}' instead."
            )
        if isinstance(self.train_set[0], dict):
            self.ununsed_features = list(set(self.train_set[0].keys()) - set(self.signature_features))
        train_iterator = DataLoader(self.train_set,
                                    sampler=RandomSampler(self.train_set),
                                    batch_size=self.opts.train_batch_size,
                                    collate_fn=self._wrap_collator_with_removed_unused_features,
                                    num_workers=self.opts.processor_num_workers)
        num_examples = len(self.train_set)
        len_iterator = len(train_iterator)
        total_train_batch_size = self.opts.train_batch_size * self.opts.gradient_accumulation_steps
        num_update_steps_per_epoch = max(len_iterator // self.opts.gradient_accumulation_steps, 1)
        if self.opts.max_steps > 0:
            max_steps = self.opts.max_steps
            num_train_epochs = math.ceil(self.opts.max_steps / num_update_steps_per_epoch)
            num_train_samples = self.opts.max_steps * total_train_batch_size
        else:
            max_steps = math.ceil(self.opts.epochs * num_update_steps_per_epoch)
            num_train_epochs = math.ceil(self.opts.epochs)
            num_train_samples = num_examples * self.opts.epochs

        # Set up learning rate scheduler
        if self.scheduler is None:
            self.create_scheduler(num_training_steps=max_steps)

        # Start fit model with trainning examples
        LOGGER.info(f"\n"
                    f"****** Summary ******\n"
                    f"  Training task                = {self.opts.task}\n"
                    f"  Training progress\n"
                    f"    Number of epochs           = {num_train_epochs}\n"
                    f"    Number of samples          = {len(self.train_set)}\n"
                    f"    Epoch optimization steps   = {len_iterator}\n"
                    f"    Global optimization steps  = {max_steps}\n"
                    f"    Train batch size           = {self.opts.train_batch_size}\n"
                    f"    Start Learning rate        = {get_lr_strings(self.optim)}\n"
                    f"  Evaluation progress\n"
                    f"    Evaluation strategy        = {self.opts.eval_steps} ({self.opts.evaluation_strategy})\n"
                    f"    Number of samples          = {num_examples}\n"
                    f"    Eval batch size            = {self.opts.eval_batch_size}\n"
                    f"    Total evaluation steps     = {len_iterator}")

        # Update Training statistics
        self.stats.set_value('max_steps', max_steps)
        self.stats.set_value('train_examples', num_examples)
        self.stats.set_value('train_epochs', num_train_epochs)
        self.stats.set_value('train_batch_size', self.opts.train_batch_size)
        self.stats.set_value('evaluation_strategy', self.opts.evaluation_strategy)

        for curr_epoch in range(1, self.opts.epochs+1):
            self.stats.set_value('curr_epoch', curr_epoch)
            trained_progress_bar = tqdm(train_iterator,
                                        total=num_update_steps_per_epoch,
                                        leave=True,
                                        position=0,
                                        desc=f'  [TRAIN] Epoch = {curr_epoch}/{num_train_epochs}; '
                                             f'Global step = {self.stats.get_value("curr_global_step")}/{max_steps}; '
                                             f'Learning rate = {get_lr_strings(self.optim)}')
            _ = self._fit_one_epoch(trained_progress_bar)

            if self.scheduler_update_strategy == 'epoch':
                self.scheduler.step()

            if self.opts.evaluation_strategy == 'epoch':
                eval_metrics = self.evaluate()
                if self._is_better_model(eval_metrics, f'eval_{self.opts.metric_for_best_model}'):
                    self._save_model()
                    self._save_training_step()

                    # Update statistics for best model
                    self.stats.set_value('best_loss', eval_metrics['eval_loss'])
                    self.stats.set_value('best_score', eval_metrics[f'eval_{self.opts.metric_for_best_model}'])
                    self.stats.set_value('best_step', self.stats.get_evaluation_step())

                    if self.early_stopping is not None:
                        self.early_stopping.restart(eval_metrics[f'eval_{self.opts.metric_for_best_model}'])

                elif self.early_stopping is not None \
                        and self.early_stopping(eval_metrics[f'eval_{self.opts.metric_for_best_model}']):
                    self.stats.set_value('closed_reason', 'Early Stopping')
                    self._close_train_progress()
                    return

        self.stats.set_value('closed_reason', 'finished')
        self._close_train_progress()
        return



from typing import Optional, List
from dataclasses import dataclass, field, asdict

from enum import Enum

import datetime

import torch


@dataclass
class TrainingOptions:
    output_dir: str = field(default=f'trainer_output_{datetime.datetime.now().strftime("%H%M%S_%d%h%Y")}',
                            metadata={'help': 'The directory where model logs, predictions and checkpoints will be '
                                              'saved.'})
    task: str = field(default=f'text-classification',
                      metadata={'help': ('Task used for training model.',
                                         'Possible values are: ',
                                         '- `"text-classification"`: ... .Default'
                                         )})
    seed: int = field(default=42,
                      metadata={'help': 'Initialization of a random seed at the start of training.'})
    no_cuda: bool = field(default=False,
                          metadata={'help': 'Do not use CUDA even when it is available'})
    no_decay_param_names: List[str] = field(default=('bias', 'LayerNorm'),
                                            metadata={'help': ''})
    # Training Hyper-parameters
    epochs: int = field(default=5,
                        metadata={'help': ''})
    learning_rate: float = field(default=5e-5,
                                 metadata={'help': ''})
    train_batch_size: int = field(default=32,
                                  metadata={'help': ''})
    eval_batch_size: int = field(default=32,
                                 metadata={'help': ''})
    metric_for_best_model: str = field(default='loss',
                                       metadata={'help': 'The metric used to compare two models in order to determine '
                                                         'the best model.'})
    greater_is_better: bool = field(default=False,
                                    metadata={'help': ('Use in conjunction with `metric_for_best_model` to specify '
                                                       'whether or not better models should have a greater metric.')})
    # Optimizer
    optimizer_name: str = field(default='adamw',
                                metadata={'help': ''})
    weight_decay: float = field(default=0.0,
                                metadata={})
    adam_beta1: float = field(default=0.9,
                              metadata={'help': ''})
    adam_beta2: float = field(default=0.999,
                              metadata={'help': ''})
    adam_epsilon: float = field(default=1e-8,
                                metadata={'help': ''})
    max_grad_norm: float = field(default=1.0,
                                 metadata={'help': ''})
    momentum: float = field(default=0,
                            metadata={'help': 'momentum factor for SGD optimizer. (default: 0)'})

    # Learning rate Scheduler
    scheduler_name: str = field(default='linear',
                                metadata={'help': ''})
    warmup_proportion: float = field(default=0.0,
                                     metadata={'help': ''})
    warmup_steps: int = field(default=0,
                              metadata={'help': ''})
    decay_step: float = field(default=1.0,
                              metadata={'help': 'Period of learning rate decay for StepLR.'})
    gamma: float = field(default=0.1,
                         metadata={'help': 'Multiplicative factor of learning rate decay for StepLR.'})
    gradient_accumulation_steps: int = field(default=1,
                                             metadata={'help': ''})
    max_steps: int = field(default=0,
                           metadata={'help': ''})
    # Processor
    processor_num_workers: int = field(default=0,
                                       metadata={'help': ''})
    # Evaluation Hyper-parameters
    evaluation_strategy: str = field(default='epoch',
                                     metadata={'help': (
                                         'The evaluation strategy to adopt during training. ',
                                         'Possible values are: ',
                                         '- `"no"`: No evaluation is done during training.',
                                         '- `"steps"`: Evaluation is done (and logged) every `eval_steps`.',
                                         '- `"epoch"`: Evaluation is done at the end of each epoch.',)})
    eval_steps: Optional[int] = field(default=1,
                                      metadata={'help': 'Number of update steps between two evaluations if '
                                                        '`evaluation_strategy="steps"`.'})
    # Early Stopping
    early_stopping_steps: int = field(default=0,
                                      metadata={'help': ''})

    early_stopping_min_delta: float = field(default=0,
                                            metadata={'help': ''})

    # Other parameters
    track_metrics: bool = field(default=False,
                                metadata={'help': 'whether to enable track various metrics such as accuracy and log '
                                                  'loss on training or validation progress.'})

    def __str__(self):
        self_as_dict = asdict(self)

        self_as_dict = {k: v for k, v in self_as_dict.items()}

        attrs_as_str = [f"{k}={v},\n" for k, v in sorted(self_as_dict.items())]
        return f"{self.__class__.__name__}(\n{''.join(attrs_as_str)})"

    __repr__ = __str__

    def __post_init__(self):
        if not self.no_cuda and torch.cuda.is_available():
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

    def to_dict(self):
        """
        Serializes this instance while replace `Enum` by their values (for JSON serialization support). It obfuscates
        the token values by removing their value.
        """
        d = asdict(self)
        for k, v in d.items():
            if isinstance(v, Enum):
                d[k] = v.value
            if isinstance(v, list) and len(v) > 0 and isinstance(v[0], Enum):
                d[k] = [x.value for x in v]
            if k.endswith("_token"):
                d[k] = f"<{k.upper()}>"
        return d



from typing import Union, List, Dict
from collections import defaultdict


class DefaultMetrics:
    @staticmethod
    def _calc_micro_metrics(tp, p, t, percent=True):
        """
            Compute overall precision, recall and FB1 (default values are 0.0)
            if percent is True, return 100 * original decimal value
        """
        micro_precision = tp / p if p else 0
        micro_recall = tp / t if t else 0
        micro_f1 = 2 * micro_precision * micro_recall / (micro_precision + micro_recall) \
            if micro_precision + micro_recall else 0

        if percent:
            return 100 * micro_precision, 100 * micro_recall, 100 * micro_f1
        else:
            return micro_precision, micro_recall, micro_f1

    def _calc_macro_metrics(self, corrects, pred, gold, l_types, percent=True):
        sum_prec, sum_rec, sum_f1 = 0.0, 0.0, 0.0
        for t in l_types:
            prec, rec, f1 = self._calc_micro_metrics(corrects[t], pred[t], gold[t], percent=False)
            sum_prec += prec
            sum_rec += rec
            sum_f1 += f1
        macro_precision, macro_recall, macro_f1 = (sum_prec / len(l_types),
                                                   sum_rec / len(l_types),
                                                   sum_f1 / len(l_types))
        if percent:
            return 100 * macro_precision, 100 * macro_recall, 100 * macro_f1
        else:
            return macro_precision, macro_recall, macro_f1

    def _calc_text_classification_score(self, true_seqs, pred_seqs, percent=True) -> Dict:
        correct_counts = defaultdict(int)
        true_counts = defaultdict(int)
        pred_counts = defaultdict(int)
        for true_tag, pred_tag in zip(true_seqs, pred_seqs):
            if true_tag == pred_tag:
                correct_counts[true_tag] += 1
            true_counts[true_tag] += 1
            pred_counts[pred_tag] += 1
        sum_correct_counts = sum(correct_counts.values())
        sum_true_counts = sum(true_counts.values())
        sum_pred_counts = sum(pred_counts.values())
        types = sorted(list(set(list(true_counts) + list(pred_counts))))
        micro_prec, micro_rec, micro_f1 = self._calc_micro_metrics(sum_correct_counts, sum_pred_counts, sum_true_counts,
                                                                   percent=True)
        macro_prec, macro_rec, macro_f1 = self._calc_macro_metrics(correct_counts, pred_counts, true_counts,
                                                                   l_types=types,
                                                                   percent=True)
        acc = sum_correct_counts/len(true_seqs)
        if percent:
            acc = acc * 100
        return {
            'accuracy': acc,
            'micro_precision': micro_prec, 'micro_recall': micro_rec, 'micro_f1': micro_f1,
            'macro_precision': macro_prec, 'macro_recall': macro_rec, 'macro_f1': macro_f1,
        }

    def __call__(self,
                 true_seqs: List[Union[str, int]],
                 pred_seqs: List[Union[str, int]],
                 task: str,
                 percent: bool = True, *args, **kwargs) -> Dict:
        if task == 'text-classification':
            return self._calc_text_classification_score(true_seqs, pred_seqs, percent)
        else:
            raise ValueError(
                f"No suitable metrics found for task `{task}` or task name not valid."
            )


import random
import numpy as np
import torch


def set_ramdom_seed(seed: int):
    if seed > 0:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        # some cudnn methods can be random even after fixing the seed
        # unless you tell it to be deterministic
        torch.backends.cudnn.deterministic = True

    if torch.cuda.is_available() and seed > 0:
        torch.cuda.manual_seed_all(seed)


def get_spend_time(start, end):
    hours, rem = divmod(end-start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)



from torch.optim.lr_scheduler import LambdaLR, StepLR


def get_linear_scheduler_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )
    return LambdaLR(optimizer, lr_lambda, last_epoch)


def get_step_scheduler_with_warmup(optimizer, step_size, gamma, last_epoch=-1):
    return StepLR(optimizer, step_size, gamma=gamma, last_epoch=last_epoch)


class SchedulerNames:
    Linear = {'name': 'linear', 'cls': get_linear_scheduler_with_warmup, 'update_strategy': 'step'}
    StepLR = {'name': 'steplr', 'cls': get_step_scheduler_with_warmup, 'update_strategy': 'epoch'}




import torch


def get_lr_strings(optimizer):
    return ",".join([f"{group['lr']:10.3e}" for group in optimizer.param_groups])


class OptimizerNames:
    AdamW = {'name': 'adamw', 'cls': torch.optim.AdamW}
    SGD = {'name': 'sgd', 'cls': torch.optim.SGD}





""" Main entry point of the Kheiron library """
from kheiron.opts import TrainingOptions

from kheiron.trainer import Trainer

__version__ = '1.0.4'
__author__ = 'TrongDat-Ngo'
__email__ = 'datnnt97@gmail.com'


__all__ = ['Trainer', 'TrainingOptions']




"""Define constant values used across Kheiron."""


class SavedFiles(object):
    WEIGHT_MODEL_FILE = 'pytorch_model.bin'
    TRAINING_ARGS_FILE = 'training_args.bin'
    TRAINING_STATS_FILE = 'training_stats.json'
    OPTIMIZER_FILE = 'optimizer.bin'
    SCHEDULER_FILE = 'scheduler.bin'



from loguru import logger

import sys


logger.remove()
logger.add(sys.stderr,
           format='<g>{time:ddd, DD MMM YYYY HH:mm:SS}</g> '
                  '| <m><b>Kheiron</b></m> '
                  '| <e><b><i>{level}</i></b></e> '
                  ': {message}',
           level='INFO')


"""Module defining various utilities."""
from kheiron.utils.logging import logger as LOGGER
from kheiron.utils.metrics import DefaultMetrics
from kheiron.utils.optimizers import OptimizerNames, get_lr_strings
from kheiron.utils.schedulers import SchedulerNames
from kheiron.utils.statistics import TrainingStats
from kheiron.utils.misc import set_ramdom_seed, get_spend_time

__all__ = ['LOGGER', 'DefaultMetrics', 'OptimizerNames', 'SchedulerNames', 'TrainingStats',
           'set_ramdom_seed',  'get_spend_time', 'get_lr_strings'
]


import pathlib
from kheiron import __version__
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()
requires = (HERE / 'requirements.txt').read_text().split('\n')

setup(
    name="Kheiron",
    version=__version__,
    description="Kheiron: Train and evaluate libary for Pytorch model.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/datnnt1997/Kheiron",
    author="DatNgo",
    author_email="datnnt97@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=('datasets', 'outputs')),
    include_package_data=True,
    install_requires=requires
)